package com.xylink.sdk;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.xylink.config.SDKConfig;
import com.xylink.config.SDKConfigMgr;
import com.xylink.model.HttpMethod;
import com.xylink.util.HttpUtil;
import com.xylink.util.Result;
import com.xylink.util.SignatureSample;
import org.apache.commons.lang3.StringUtils;

import java.io.IOException;
import java.util.Map;
import java.util.TreeMap;

import static com.xylink.config.SignatureVersion.Signature_2_0_1;

/**
 * @Classname SdkMethod
 * @Date 2022/4/6 10:39 上午
 * @Created by anan
 * @Description sdkjar 通用方法，用于用户调用SDKserver的接口，用户需要传入url，body以及选择正确的请求方法即可
 * url的形式如下：/api/rest/xxx?xxx=xxxxx
 * 需要事先进行初始化
 *
 * #####################
 * 需要使用签名算法1.0   #
 * ####################
 */
public class SdkMethodV1 {
    private static final SignatureSample SIGNATURE_SAMPLE = new SignatureSample();

    /**
     * 使用签名方法1.0
     * @param url   /api/rest/xxx?xxx=xxxxx,不包含&signature=XXX
     * @param token 企业token
     * @return 返回文档描述中的对象
     * @throws IOException 抛出异常
     */
    public static <T> Result<T> sdkGet(String url, String token, Class<T> responseType) throws IOException {
        url = appendSignAndHost(url, token, StringUtils.EMPTY, HttpMethod.GET.name());
        return HttpUtil.getResponse(url, HttpMethod.GET.name(), StringUtils.EMPTY, responseType);
    }

    /**
     * @param url   /api/rest/xxx?xxx=xxxxx,不包含&signature=XXX
     * @param token 企业token
     * @param body  文档描述中的body内容
     * @return 返回文档描述中的对象
     * @throws IOException 抛出异常
     */
    public static <T> Result<T> sdkPost(String url, String token, Object body, Class<T> responseType) throws IOException {
        //body序列化
        String jsonEntity = new ObjectMapper().writeValueAsString(body);
        url = appendSignAndHost(url, token, jsonEntity, HttpMethod.POST.name());
        return HttpUtil.getResponse(url, HttpMethod.POST.name(), jsonEntity, responseType);
    }

    /**
     * @param url   /api/rest/xxx?xxx=xxxxx,不包含&signature=XXX
     * @param token 企业token
     * @param body  文档描述中的body内容
     * @return 返回文档描述中的对象
     * @throws IOException 抛出异常
     */
    public static <T> Result<T> sdkPut(String url, String token, Object body, Class<T> responseType) throws IOException {
        //body序列化
        String jsonEntity = new ObjectMapper().writeValueAsString(body);
        url = appendSignAndHost(url, token, jsonEntity, HttpMethod.PUT.name());
        return HttpUtil.getResponse(url, HttpMethod.PUT.name(), jsonEntity, responseType);
    }

    /**
     * @param url   /api/rest/xxx?xxx=xxxxx,不包含&signature=XXX
     * @param token 企业token
     * @return 返回文档描述中的对象
     * @throws IOException 抛出异常
     */
    public static <T> Result<T> sdkDelete(String url, String token, Class<T> responseType) throws IOException {
        url = appendSignAndHost(url, token, StringUtils.EMPTY, HttpMethod.DELETE.name());
        return HttpUtil.getResponse(url, HttpMethod.DELETE.name(), StringUtils.EMPTY, responseType);
    }

    /**
     * 添加签名
     *
     * @param url    url
     * @param token  token
     * @param body   body
     * @param method 请求方法
     * @return 返回拼接后的url
     */
    private static String appendSignAndHost(String url, String token, String body, String method) {
        String signature = SIGNATURE_SAMPLE.computeSignature(body, method, token, url);
        url += "&signature=" + signature;
        //2.0.1版本的签名需要规范化url

        SDKConfig sdkConfig = SDKConfigMgr.getSdkConfig(url);
        if (sdkConfig.getSignatureVersion().equals(Signature_2_0_1)){
            url = canonicalizedUri(url);
        }
        url = sdkConfig.getServerHost() + url;
        return url;
    }



    /**
     * 规范化uri
     * @param uri 原始请求uri (以/开始的完整请求地址, 包含)
     * @return 规范化后的uri
     * eg: /api?b=2&a=1&A=0
     * return: /api?A=0&a=1&b=2
     */
    private static String canonicalizedUri(String uri) {
        final int index = uri.indexOf("?");
        if (index > -1) {
            String uriPrefix = uri.substring(0, index);
            String uriSuffix = uri.substring(index + 1);
            StringBuilder canonicalizedUri = new StringBuilder(uriPrefix);
            StringBuilder canonicalizedQueryString = new StringBuilder();
            TreeMap<String, String> queryMap = new TreeMap<String, String>();

            if (uriSuffix.length() > 0) {
                final String[] pairs = uriSuffix.split("&");
                for (String pair : pairs) {
                    final int idx = pair.indexOf("=");
                    final String key = idx > 0 ? pair.substring(0, idx) : pair;
                    if (key.length() > 0 && !queryMap.containsKey(key)) {
                        final String value = idx > 0 ? pair.substring(idx + 1) : "";
                        queryMap.put(key, value);
                    }
                }
                for (Map.Entry<String, String> entry : queryMap.entrySet()) {
                    canonicalizedQueryString.append(entry.getKey()).append("=").append(entry.getValue()).append("&");
                }
                if (canonicalizedQueryString.length() > 1) {
                    canonicalizedQueryString.setLength(canonicalizedQueryString.length() - 1);
                }
            }
            if (canonicalizedQueryString.length() > 0) {
                canonicalizedUri.append("?").append(canonicalizedQueryString);
            }
            return canonicalizedUri.toString();
        }
        return uri;
    }

}
